import os
import time
import types
from typing import Any, Dict, List

import ulid

from ..serialize import get_content, get_request_body


class DjangoFilter:
    use_frames_of_interest = False

    def __init__(self, config) -> None:
        self.config = config
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)
        self.data: Dict[str, Any] = {}
        self.timestamp: float

    def __call__(self, frame: types.FrameType, event: str, arg: object) -> bool:
        co_name = frame.f_code.co_name
        filename = frame.f_code.co_filename
        return (
            self.use_frame_boundaries
            and co_name == "get_response"
            and os.path.normpath("/kolo/middleware.py") in filename
        )

    def process(
        self,
        frame: types.FrameType,
        event: str,
        arg: object,
        call_frame_ids: List[Dict[str, str]],
    ):
        if event == "call":
            self.timestamp = time.time()
            request = frame.f_locals["request"]
            self.request_data = {
                "frame_id": f"frm_{ulid.new()}",
                "scheme": request.scheme,
                "method": request.method,
                "path_info": request.path_info,
                "body": get_request_body(request),
                "headers": dict(request.headers),
                "url_pattern": None,
                "type": "django_request",
            }
            return self.request_data
        elif event == "return":  # pragma: no branch
            duration = time.time() - self.timestamp
            ms_duration = round(duration * 1000, 2)

            request = frame.f_locals["request"]
            match = request.resolver_match
            # match can be None before url resolving
            assert match
            self.request_data["url_pattern"] = {
                "namespace": match.namespace,
                "route": match.route,
                "url_name": match.url_name,
                "view_qualname": match._func_path,
            }

            response = frame.f_locals["response"]
            return {
                "frame_id": f"frm_{ulid.new()}",
                "ms_duration": ms_duration,
                "status_code": response.status_code,
                "content": get_content(response),
                "headers": dict(response.items()),
                "type": "django_response",
            }



import logging
import os
import sys
import threading
from typing import Any, Callable, Dict, Tuple

from django.conf import settings
from django.http import HttpRequest, HttpResponse

from .config import create_kolo_directory, load_config_from_toml
from .db import load_config_from_db, setup_db
from .profiler import KoloProfiler

logger = logging.getLogger("kolo")

DjangoView = Callable[[HttpRequest], HttpResponse]


class KoloMiddleware:
    def __init__(self, get_response: DjangoView) -> None:
        self._get_response = get_response
        self.enabled = self.should_enable()
        if self.enabled:
            kolo_directory = create_kolo_directory()
            self.config = load_config_from_toml(kolo_directory / "config.toml")
            self.db_path = setup_db(self.config)

    def __call__(self, request: HttpRequest) -> HttpResponse:
        if not self.enabled or self.check_for_third_party_profiler():
            return self.get_response(request)

        self.config = load_config_from_db(self.db_path, self.config)
        self.use_frame_boundaries = self.config.get("use_frame_boundaries", False)

        filter_config = self.config.get("filters", {})
        ignore_request_paths = filter_config.get("ignore_request_paths", [])
        for path in ignore_request_paths:
            if path in request.path:
                return self.get_response(request)

        self.profiler = KoloProfiler(self.db_path, config=self.config)
        if not self.use_frame_boundaries:
            self.profiler.initialize_request(request)

        with self.profiler:
            response = self.get_response(request)

        if not self.use_frame_boundaries:
            self.profiler.finalize_response(response)

        threading.Thread(
            target=self.profiler.save_request_in_db, name="kolo-save_request_in_db"
        ).start()

        return response

    def process_view(
        self,
        request: HttpRequest,
        view_func: DjangoView,
        view_args: Tuple[Any],
        view_kwargs: Dict[str, Any],
    ):
        if self.enabled and not self.use_frame_boundaries:
            self.profiler.set_url_pattern(request)

    def get_response(self, request: HttpRequest) -> HttpResponse:
        response = self._get_response(request)
        return response

    def check_for_third_party_profiler(self) -> bool:
        profiler = sys.getprofile()
        if profiler:
            logger.warning("Profiler %s is active, disabling KoloMiddleware", profiler)
            return True
        return False

    def should_enable(self) -> bool:
        if settings.DEBUG is False:
            logger.debug("DEBUG mode is off, disabling KoloMiddleware")
            return False

        if os.environ.get("KOLO_DISABLE", "false").lower() not in ["false", "0"]:
            logger.debug("KOLO_DISABLE is set, disabling KoloMiddleware")
            return False

        if self.check_for_third_party_profiler():
            return False

        return True
